'==========================================================================
'
' VBScript Source File -- Created with SAPIEN Technologies PrimalScript 2007
'
' NAME: Group_Shadow.vbs
'
' AUTHOR: Dan Holme , Intelliem
' DATE  : 12/12/2007
'
' USAGE:  
' cscript.exe Group_Shadow.vbs
'
' Dynamically updates the membership of a group
' to match the objects returned from an Active Directory query
'
' See the Windows Administration Resource Kit for documentation
'
' Neither Microsoft nor Intelliem guarantee the performance
' of scripts, scripting examples or tools.
'
' See www.intelliem.com/resourcekit for updates to this script
'
' (c) 2007 Intelliem, Inc
'==========================================================================

Option Explicit

Dim sDomainDN
Dim sGroupSAMAccountName
Dim aSearchOUs
Dim sQuery

'==========================================================================
' CONFIGURATION BLOCK
' Domain's DN
sDomainDN = "dc=contoso,dc=com"
' sAMAccountName of shadow group
sGroupSAMAccountName = "Employees"
' An array of one or more OUs to search
aSearchOUs = Array("ou=employees,ou=people,dc=contoso,dc=com", _
				   "ou=contractors,ou=people,dc=contoso,dc=com")
' LDAP query that will be run in each OU
sQuery = "(&(objectCategory=person)(objectClass=user));distinguishedName;subtree"
'==========================================================================

' Create dictionaries
Dim dResults
Set dResults = CreateObject("Scripting.Dictionary")
dResults.CompareMode = vbTextCompare ' Case INsensitive
Dim dTargetMembership
Set dTargetMembership = CreateObject("Scripting.Dictionary")
dTargetMembership.CompareMode = vbTextCompare ' Case INsensitive
Dim dCurrentMembership
Set dCurrentMembership = CreateObject("Scripting.Dictionary")
dCurrentMembership.CompareMode = vbTextCompare ' Case INsensitive
Dim dMembershipChanges
Set dMembershipChanges = CreateObject("Scripting.Dictionary")
dMembershipChanges.CompareMode = vbTextCompare ' Case INsensitive

' Perform LDAP searches, adding to final list stored in dTargetMembership
Dim sSearchOU
Dim sLDAPQuery
For Each sSearchOU In aSearchOUs
	sLDAPQuery = "<LDAP://" & sSearchOU & ">;" & sQuery
	Set dResults = AD_Search_Dictionary(sLDAPQuery)
	Call DictionaryAppend(dResults, dTargetMembership)
Next

' Locate group
Dim sGroupADsPath
Dim oGroup
sGroupADsPath = ADObject_Find_Generic(sGroupSAMAccountName, sDomainDN)
If sGroupADsPath = "" Then
	' Error handling: group not found
	WScript.Quit
End If
Set oGroup = GetObject(sGroupADsPath)

' Get members and store in dictionary
Dim aMembers
aMembers = oGroup.GetEx("member")
Set dCurrentMembership = ArrayToDictionary(aMembers)

' Calculate the "delta" between the current and desired state
Set dMembershipChanges = Dictionary_Transform(dCurrentMembership, dTargetMembership)

' Make the membership changes based on the transform dictionary's instructions
Dim sMember
For Each sMember In dMembershipChanges
	If UCase(dMembershipChanges.Item(sMember)) = "ADD" Then
		oGroup.Add "LDAP://" & sMember
	End If
	If UCase(dMembershipChanges.Item(sMember)) = "DELETE" Then
		oGroup.Remove "LDAP://" & sMember
	End If
Next

WScript.Quit

' ======================
' FUNCTIONS FROM LIBRARY
' ======================

' #region Dictionary routines

Function ArrayToDictionary(ByRef aArray)
	' Converts a one-dimensional array into a dictionary.
	' Assumes elements in array are unique
	Dim dDic
	Dim aElement
	Set dDic = CreateObject("Scripting.Dictionary")
	dDic.CompareMode = vbTextCompare ' Case INsensitive
	
	On Error Resume Next ' trap duplicate array elements
	For Each aElement In aArray
		dDic.Add aElement, 0		
	Next
	On Error GoTo 0
	
	Set ArrayToDictionary = dDic
End Function

Sub DictionaryAppend(ByRef dNewElements, ByRef dDictionary)
	' Appends the elements of dNewElements to dDictionary
	Dim sKey
	
	On Error Resume Next ' trap duplicate array elements
	For Each sKey In dNewElements.keys
		dDictionary.Add sKey, dNewElements.Item(sKey)
	Next
	On Error GoTo 0
End Sub

Function Dictionary_Transform(ByVal dOriginal, ByVal dFinal)
	' Retunrs a dictionary with a list of update operations required
	' so that dOriginal is transformed to dFinal
	
	Dim dTransform, sKey
	Set dTransform = CreateObject("Scripting.Dictionary")
	dTransform.CompareMode = vbTextCompare ' Case INsensitive
	
	For Each sKey In dFinal.Keys
		If Not dOriginal.Exists(sKey) Then
			dTransform.Add sKey, "ADD"
		End If
	Next
	
	For Each sKey In dOriginal.Keys
		If Not dFinal.Exists(sKey) Then
			dTransform.Add sKey, "DELETE"
		End If
	Next
	
	Set Dictionary_Transform = dTransform
	
End Function

' #endregion

' #region Active Directory object find routines

Function ADObject_Find_Generic(ByVal sObject, ByVal sSearchDN)
	' Version 071130
	' Takes any input (name, DN, or ADsPath) of a user, computer, or group, and
	' returns the ADsPath of the object as a way of validating that the object exists
	'
	' INPUTS:	sObject 				DN or ADsPath to an object
	'					 				sAMAccountName (pre-Windows 2000 logon name) of a user or group
	'					 				computer name of a computer
	'			sSearchDN 				the DN within which to search (often, the DN of the domain, e.g. dc=contoso, dc=com)
	'
	' RETURNS:	ADObject_Find_Generic 	ADsPath (LDAP://...) of the object
	'								   	blank if object was not found
	'
	' NOTES:	ASSUMPTION: computers, users & groups have unique names. See note inline.
	'
	' REQUIRES	AD_Search_Array routine
	'			AD_Search_RS routine
	' 			ADObject_Validate routine
	
	Dim aResults, sLDAPQuery
	Select Case ADObject_NameType(sObject)
		Case ""
			ADObject_Find_Generic = ""
		Case "adspath"
			ADObject_Find_Generic = ADObject_Validate(sObject)
		Case "distinguishedname"
			ADObject_Find_Generic = ADObject_Validate("LDAP://" & sObject)
		Case "name"
				' Assumption: No computer has the same name as a user's or group's sAMAccountName
				' otherwise, this query will return more than one result
				sLDAPQuery = "<LDAP://" & sSearchDN & ">;" & _
							 "(|(samAccountName=" & sObject & ")(samAccountName=" & sObject & "$));" & _
							 "aDSPath;subtree"
				aResults = AD_Search_Array (sLDAPQuery)
				If Ubound(aResults) = -1 Then
					ADObject_Find_Generic = ""
				Else
					ADObject_Find_Generic = aResults(0)
				End If
	End Select
	
End Function

Function ADObject_NameType(ByVal sObjectName)
	' Version 071204
	' Evaluates sObjectName to determine what type of name it is
	' Returns	ADObject_NameType	adspath
	'								distinguishedname
	'								name
	'								blank if sObjectName = ""
	
	Dim sNameType
	
	If Len(sObjectName) = 0 Then
		sNameType = ""

	ElseIf Len(sObjectName) < 3 Then
		' can't be a DN or an ADsPath - must be a name
		sNameType = "name"
	
	ElseIf Ucase(Left(sObjectName,3)) = "CN=" Then
		' is a DN
		sNameType = "distinguishedname"
		
	ElseIf Len(sObjectName) < 8 Then
		' too short to be an ADsPath and isn't a DN, so it must be a name
		sNameType = "name"
		
	ElseIf UCase(Left(sObjectName, 7)) = "LDAP://" Then
		' is already an ADsPath
		sNameType = "adspath"

	Else
		' must be a name
		sNameType = "name"

	End If

	ADObject_NameType = sNameType
End Function

Function ADObject_Validate(ByVal sObjectADsPath)
	' Version 071122
	' Returns ADsPath of object as a way of validating that the object exists
	'
	' INPUTS:	sObjectADsPath		ADsPath of object to test
	' RETURNS:	ADObject_Validate	Path of object (if it exists) or blank
	
	Dim oObject
	On Error Resume Next
	Set oObject = GetObject(sObjectADsPath)
	If Err.Number <> 0 Then
		ADObject_Validate = ""
		Err.Clear
	Else
		ADObject_Validate = oObject.ADsPath
	End If	
End Function

Function AD_Search_Dictionary(sLDAPQuery)
	' Version 071204
	' Returns a dictionary of AD objects
	' A wrapper around AD_Search_RS
	' Inputs:	sLDAPQuery				The query to perform
	' Returns	AD_Search_Dictionary	Dictionary with KEY set to the first
	'									attribute returned by the search, and
	'									the item set to the second.
	'									The sLDAPQuery should specify a unique
	'									attribute (e.g. DN or ADsPath) for the
	'									first attribute!
	
	Dim rs, dResults, sKey, sItem
	Set dResults = CreateObject("Scripting.Dictionary")
	dResults.CompareMode = vbTextCompare ' Case INsensitive

	Set rs = AD_Search_RS(sLDAPQuery)
	If Not(rs Is Nothing) Then
		If Not (rs.BOF And rs.EOF) Then
			rs.MoveFirst
			Do While Not rs.EOF
				sKey = rs.Fields(0)
				sItem = ""
				On Error Resume Next ' Trap in case there is no second field
				sItem = rs.Fields(1)
				On Error GoTo 0
				dResults.Add sKey, sItem
				rs.MoveNext
			Loop
		End If
	Else
		' Handle error: Query returned no results
	End If
	Set AD_Search_Dictionary = dResults
	
End Function

Function AD_Search_Array(sLDAPQuery)
	' Version 071130
	' Returns an array of AD objects
	' A wrapper around AD_Search_RS
	' Inputs:	sLDAPQuery		The query to perform
	' Returns	AD_Search_Array	Array of Field(0) of recordset
	'							or empty array if recordset is empty
	
	Dim rs, aResults
	aResults = array()
	Set rs = AD_Search_RS(sLDAPQuery)
	If Not(rs Is Nothing) Then
		If Not (rs.BOF And rs.EOF) Then
			rs.MoveFirst
			Do While Not rs.EOF
				Redim Preserve aResults(ubound(aResults) + 1)
				aResults(ubound(aResults)) = rs.Fields(0)
				rs.MoveNext
			Loop
		End If
	Else
		' Error handling: query returned Nothing
	End If
	AD_Search_Array = aResults
	
End Function

Function AD_Search_RS(sLDAPQuery)
	' Version 071130
	' Returns a recordset of AD objects
	' Inputs:	sLDAPQuery		The query to perform
	' Returns:	AD_Search_RS	Recordset
	
	Dim oConnection
	
	' Open an ADO connection using null credentials
	Set oConnection = CreateObject("ADODB.Connection")
	oConnection.Provider = "ADsDSOObject"
	oConnection.Open "", vbNullString, vbNullString
	If oConnection.State = 0 Then ' 0 = adStateClosed
	     ' Error handling code: can't connect to AD
		Set AD_Search_RS = Nothing
	Else
		Set AD_Search_RS = oConnection.Execute(sLDAPQuery)
	End If
End Function

' #endregion